#######################################
# All this code comes from fastai/nbdev
#######################################

from nbconvert.preprocessors import ExecutePreprocessor
from nbdev.export import _mk_flag_re

TEST_FLAGS = "slow"


class _ReTstFlags:  # pragma: no cover
    "Test flag matching regular expressions"

    def __init__(self, all_flag):
        "match flags applied to all cells?"
        self.all_flag = all_flag

    def _deferred_init(self):
        "Compile at first use but not before since patterns need `get_config().tst_flags`"
        if hasattr(self, "_re"):
            return
        tst_flags = TEST_FLAGS
        tst_flags += f"|skip" if tst_flags else "skip"
        _re_all = "all_" if self.all_flag else ""
        self._re = _mk_flag_re(f"{_re_all}({tst_flags})", 0, "Any line with a test flag")

    def findall(self, source):
        self._deferred_init()
        return self._re.findall(source)

    def search(self, source):
        self._deferred_init()
        return self._re.search(source)


_re_all_flag = _ReTstFlags(True)


def get_all_flags(cells):
    "Check for all test flags in `cells`"
    result = []
    for cell in cells:
        if cell["cell_type"] == "code":
            result.extend(_re_all_flag.findall(cell["source"]))
    return set(result)


_re_flags = _ReTstFlags(False)


def get_cell_flags(cell):
    "Check for any special test flag in `cell`"
    if cell["cell_type"] != "code" or len(TEST_FLAGS) == 0:
        return []
    return _re_flags.findall(cell["source"])


class NoExportPreprocessor(ExecutePreprocessor):
    "An `ExecutePreprocessor` that executes cells that don't have a flag in `flags`"

    def __init__(self, flags, pip_install=True, **kwargs):
        self.flags = flags
        self.pip_install = pip_install
        super().__init__(**kwargs)

    def preprocess_cell(self, cell, resources, index):
        if "source" not in cell or cell["cell_type"] != "code":
            return cell, resources
        for f in get_cell_flags(cell):
            if f not in self.flags:
                return cell, resources
        if "!pip" in "".join(cell["source"]) and not self.pip_install:
            code = cell["source"]
            print(f"Skipping: {code}")
            return cell, resources

        # if check_re(cell, _re_notebook2script): return cell, resources
        return super().preprocess_cell(cell, resources, index)



import time
from pathlib import Path
from types import SimpleNamespace
from typing import Union, List
from tempfile import TemporaryDirectory
from fastcore.basics import listify

import nbformat
from fastcore.script import *
from rich import print as pprint
from rich.progress import track

from nb_helpers.utils import find_nbs, search_string_in_nb, read_nb, RichLogger
from nb_helpers.colab import get_colab_url
from nb_helpers.nbdev_test import NoExportPreprocessor, get_all_flags

FEATURES = ["Path", "os", "chain", "Union", "sleep"]
## ["WandbCallback", "WandbLogger", "Table", "Artifact", "sweep"]

__all__ = ["run_one", "run_nbs"]


STATUS = SimpleNamespace(
    ok="[green]Ok[/green]:heavy_check_mark:", fail="[red]Fail[/red]", skip="[green]Skipped[/green]:heavy_check_mark:"
)


def _format_row(fname: Path, status: str, time: str, xtra_col=None, fname_only: bool = True) -> tuple:
    "Format one row for a rich.Table"

    formatted_status = getattr(STATUS, status.lower())
    fname = fname.name if fname_only else fname
    row = (str(fname), formatted_status, f"{int(time)}s")
    if len(listify(xtra_col)) > 0:
        row += (str(xtra_col),)
    return row


def skip_nb(notebook, flags=None, filters=None):
    "check for notebook flags: all_skip, all_slow and filters: tensorflow, pytorch, ..."
    skip = False
    for f in get_all_flags(notebook["cells"]):
        if f not in listify(flags):
            skip = True
    if not search_string_in_nb(notebook, filters):
        skip = True
    return skip


def exec_nb(notebook, flags=None, timeout=600, use_temp_dir=True, pip_install=True):
    "run notebook, possible skiping cells with flags"
    processor = NoExportPreprocessor(flags, pip_install=pip_install, timeout=timeout, kernel_name="python3")
    processed_nb = nbformat.from_dict(notebook)
    with TemporaryDirectory() as temp_dir:
        resources = {"metadata": {"path": temp_dir}} if use_temp_dir else None
        processor.preprocess(processed_nb, resources)
    return True


def run_one(
    fname: Union[Path, str],
    verbose: bool = False,
    timeout: int = 600,
    flags: List[str] = None,
    lib_name: str = None,
    no_run: bool = False,
    pip_install=True,
):
    "Run nb `fname` and timeit, recover exception"
    start = time.time()
    did_run, skip, error = False, False, None
    flags = listify(flags)
    try:
        # read notebook as dict
        notebook = read_nb(fname)

        # check if notebooks has to be runned
        skip = skip_nb(notebook, flags, lib_name)

        if skip or no_run:
            return _format_row(fname, "skip", time.time() - start), None
        else:
            did_run = exec_nb(notebook, flags, timeout, pip_install=pip_install)
    except Exception as e:
        if verbose:
            print(f"\nError in {fname}:\n{e}")
            error = e
        error = e
    return (
        _format_row(fname, "ok" if did_run else "fail", time.time() - start),
        error,
    )


@call_parse
def run_nbs(
    path: Param("A path to nb files", str) = ".",
    verbose: Param("Print errors along the way", store_true) = False,
    flags: Param("Space separated list of flags", str) = None,
    timeout: Param("Max runtime for each notebook, in seconds", int) = 600,
    lib_name: Param("Python lib names to filter, eg: tensorflow", str) = None,
    no_run: Param("Do not run any notebook", store_true) = False,
    post_issue: Param("Post the failure in github", store_true) = False,
    no_install: Param("Do not install anything with pip", store_false) = False,
):
    logger = RichLogger(columns=["fname", "status", "t[s]"])
    path = Path(path)
    files = find_nbs(path)

    failed_nbs = {}
    for nb_path in track(files, description="Running nbs..."):
        (fname, run_status, runtime), e = run_one(
            nb_path,
            verbose=verbose,
            timeout=timeout,
            flags=flags,
            lib_name=lib_name,
            no_run=no_run,
            pip_install=no_install,
        )
        pprint(f" > {fname:80} | {run_status:40} | {runtime:5} ")
        logger.writerow([fname, run_status, runtime], colab_link=get_colab_url(nb_path))
        time.sleep(0.1)
        if e is not None:
            failed_nbs[str(nb_path)] = e

    logger.to_table()
    logger.to_md("run.md")
    return failed_nbs



from tests import TEST_NB, FAIL_NB, SKIP_NB, FLAG_NB, TEST_PATH
from nb_helpers.run import STATUS, run_nbs, run_one


_STATUS = {
    "[green]Ok[/green]:heavy_check_mark:": "ok",
    "[red]Fail[/red]": "fail",
    "[green]Skipped[/green]:heavy_check_mark:": "skip",
}


def test_run_one():
    # run and pass
    row, no_errors = run_one(TEST_NB)
    assert no_errors is None, f"Notebook {TEST_NB} failed"
    assert row[1] == STATUS.ok, f"Error, the notebooks produces {_STATUS[row[1]]}!=ok"

    # run and fail
    row, errors = run_one(FAIL_NB)
    assert errors is not None, f"Notebook {FAIL_NB} has no errors, it should!"
    assert row[1] == STATUS.fail, f"Error, the notebooks produces {_STATUS[row[1]]}!=fail"

    # s kipped
    row, no_errors = run_one(SKIP_NB)
    assert no_errors is None, f"Notebook {TEST_NB} failed"
    assert row[1] == STATUS.skip, f"Error, the notebooks produces {_STATUS[row[1]]}!=skip"

    # skipped cause pandas is not used in the nb
    row, _ = run_one(TEST_NB, lib_name="pandas")
    assert row[1] == STATUS.skip, f"Error, the notebooks produces {_STATUS[row[1]]}!=skip"

    # run cause pathlib is there
    row, _ = run_one(TEST_NB, lib_name="pathlib")
    assert row[1] == STATUS.ok, f"Error, the notebooks produces {_STATUS[row[1]]}!=ok"

    # should not run cause it's flagged as slow
    row, _ = run_one(FLAG_NB)
    assert row[1] == STATUS.skip, f"Error, the notebooks produces {_STATUS[row[1]]}!=skip"

    # should run cause it's flagged as slow
    row, _ = run_one(FLAG_NB, flags="slow")
    assert row[1] == STATUS.ok, f"Error, the notebooks produces {_STATUS[row[1]]}!=ok"



import io, json, sys
from pathlib import Path

from rich import print
from rich.console import Console
from rich.table import Table
from fastcore.script import *
from fastprogress import progress_bar

from nbdev.clean import clean_nb
from nb_helpers.utils import print_output, is_nb, find_nbs


def _create_table():
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Notebook Path", style="dim")
    table.add_column("Status")
    return table


CONSOLE = Console()
CLEAN_TABLE = _create_table()


def clean_one(fname: Path, clear_outs: bool = False, disp: bool = False):
    """Clean notebook metadata:
    - `clear_all` removes also outputs
    - `disp` prints to stdout
    """
    if not is_nb(fname):
        print(f"This {fname}: is not a notebook my friend")
        return
    notebook = json.load(open(str(fname), "r", encoding="utf-8"))
    clean_nb(notebook, clear_all=clear_outs)
    if disp:
        print_output(notebook)
    else:
        x = json.dumps(notebook, sort_keys=True, indent=1, ensure_ascii=False)
        with io.open(fname, "w", encoding="utf-8") as f:
            f.write(x)
            f.write("\n")


def clean_all(path: Path, clear_outs=True, disp=False):
    "Apply clean to all nbs inside path recursvely"

    for nb in progress_bar(find_nbs(path), leave=False):
        try:
            clean_one(nb, clear_outs, disp)
            CLEAN_TABLE.add_row(str(nb), "[green]Ok[/green]:heavy_check_mark:")
        except:
            CLEAN_TABLE.add_row(str(nb), "[red]Failed[/red]")


@call_parse
def clean_nbs(
    path: Param("The path to notebooks", str) = ".",
    clear_outs: Param("Remove cell outputs", store_true) = False,
    verbose: Param("Rnun on verbose mdoe", store_true) = False,
):
    "Clean notebooks on `path` from useless metadata"
    path = Path(path)
    clean_all(path, clear_outs, disp=verbose)
    CONSOLE.print(CLEAN_TABLE)
